@@ -2,12 +2,13 @@ from __future__ import annotations
 
 import os
 import tempfile
-from typing import Callable, Optional, Tuple, Union
+from typing import Callable, List, Optional, Tuple, Union
 
 import clip
 import gdown
 import torch
 import torch.nn.functional as F
+from clip.model import CLIP
 from PIL import Image
 from pytorch_lightning import LightningModule
 from torch import Tensor, optim
@@ -123,36 +124,83 @@ class ClipDecoderInferenceModel:
 
     @torch.cuda.amp.autocast()
     @torch.no_grad()
-    def __call__(
-        self, x: Tensor, max_len: int = 64, temperature: float = 1e-8, topk: int = 1
-    ) -> str:
-        embedding_size = x.size(-1)
-        encoder_hidden_states = x.reshape(1, -1, embedding_size).to(self.device)
-        input_ids = torch.tensor(
-            self.tokenizer.bos_token_id,
-            device=self.device,
-        ).reshape(1, -1)
+    def __call__(self, x: Tensor, max_len: int = 64, beam_size: int = 1) -> str:
+        """Inference using beam search. For beam search, we predict one token per step.
+        After each step, we keep only the 'beam_size' output sequences with the highest
+        end-to-end confidence score. Repeat this process until at most 'max_len' tokens
+        have been generated.
+        """
+        encoder_hidden_states = x.reshape(1, 1, -1).to(self.device)
+        # Since we haven't performed any beam search steps yet, we just have one
+        # set of input IDs (with a single "start" token). We use 'None' for the log
+        # probability of this sequence, since it's not being predicted by the model.
+        input_ids = [torch.tensor([self.tokenizer.bos_token_id], device=self.device)]
+        beam_logprobs: Optional[List[float]] = None
+
+        def _get_beam_outputs(_input_ids: Tensor) -> Tuple[List[Tensor], Tensor]:
+            """Performs inference on the 'input_ids' Tensor, and collects the top
+            'beam_size' results by score. Returns a list of output Tensors, and
+            their respective log-probabilities.
+            """
+            outputs = self.model(_input_ids.unsqueeze(0), encoder_hidden_states)
+            logits: Tensor = outputs.logits[0, -1]
+            logprobs = F.log_softmax(logits, dim=-1)
+
+            topk_logprobs = logprobs.topk(k=beam_size)
+            indices = topk_logprobs.indices
+            logprobs = topk_logprobs.values
+            output_ids = [
+                torch.cat([_input_ids, idx.reshape(-1)], dim=0) for idx in indices
+            ]
+
+            return output_ids, logprobs
 
         for _ in range(max_len - 1):
-            outputs = self.model(input_ids, encoder_hidden_states)
-            logits: Tensor = outputs.logits[0, -1]
+            output_ids: List[Tensor] = []
+            logprobs: List[float] = []
+            beams_done: List[bool] = []
 
-            topk_logits = logits.topk(k=topk, dim=-1)
-            probs = F.softmax(topk_logits.values / temperature, dim=-1)
-            idx = torch.argmax(probs * torch.rand_like(probs))
-            pred = topk_logits.indices[idx]
-            if pred.item() == self.tokenizer.eos_token_id:
+            # Collect the top 'beam_size' results from each beam individually.
+            for beam_idx, ids in enumerate(input_ids):
+                # If 'beam_logprobs' is already defined, then we've predicted at least
+                # one token already. And if the last token is equal to the "stop" token,
+                # we don't need to perform inference with this beam anymore.
+                if beam_logprobs and ids[-1].item() == self.tokenizer.eos_token_id:
+                    output_ids.append(ids)
+                    logprobs.append(beam_logprobs[beam_idx])
+                    beams_done.append(True)
+                    continue
+
+                _output_ids, _logprobs = _get_beam_outputs(ids)
+                if beam_logprobs is not None:
+                    # Sum the log-probabilities of the existing beam and our predicted
+                    # token to get the total log-probability.
+                    _logprobs += beam_logprobs[beam_idx]
+
+                # Append the results from this beam to the aggregate lists.
+                output_ids += _output_ids
+                logprobs += _logprobs.tolist()
+                beams_done.append(False)
+
+            if all(beams_done):
+                # All search beams are done generating text.
                 break
 
-            input_ids = torch.cat([input_ids, pred.reshape(1, 1)], dim=1)
+            # Keep only the top 'beam_size' beams by total log-probability.
+            indices = torch.tensor(logprobs).topk(k=beam_size).indices
+            input_ids = [output_ids[idx] for idx in indices]
+            beam_logprobs = [logprobs[idx] for idx in indices]
 
-        return self.tokenizer.decode(input_ids.flatten(), skip_special_tokens=True)
+        # Find the predicted beam with highest overall log-probability.
+        best_beam_idx: int = torch.tensor(beam_logprobs).argmax().item()
+        # Decode the predicted token IDs into a text string.
+        return self.tokenizer.decode(input_ids[best_beam_idx], skip_special_tokens=True)
 
 
 class ImageCaptionInferenceModel(ClipDecoderInferenceModel):
     def __init__(self, model: ClipDecoder, tokenizer: GPT2Tokenizer):
         super().__init__(model, tokenizer)
-        self._clip_model: Optional[torch.nn.Module] = None
+        self._clip_model: Optional[CLIP] = None
         self._clip_preprocessor: Optional[Callable] = None
 
     def _load_clip(self):
@@ -161,7 +209,7 @@ class ImageCaptionInferenceModel(ClipDecoderInferenceModel):
         )
 
     @property
-    def clip_model(self) -> torch.nn.Module:
+    def clip_model(self) -> CLIP:
         if self._clip_model is None:
             self._load_clip()
         assert self._clip_model is not None, "Could not load CLIP model."
@@ -180,14 +228,11 @@ class ImageCaptionInferenceModel(ClipDecoderInferenceModel):
         self,
         image: Union[str, Image.Image],
         max_len: int = 64,
-        temperature: float = 1e-8,
-        topk: int = 1,
+        beam_size: int = 1,
     ) -> str:
         if isinstance(image, str):
             image = Image.open(image)
 
-        preprocessed = self.clip_preprocessor(image).to(self.device)
+        preprocessed: Tensor = self.clip_preprocessor(image).to(self.device)
         encoded = self.clip_model.encode_image(preprocessed.unsqueeze(0))
-        return super().__call__(
-            encoded, max_len=max_len, temperature=temperature, topk=topk
-        )
+        return super().__call__(encoded, max_len=max_len, beam_size=beam_size)
